{% extends "cb_base.html" %}

{% block content %}

   <p>
      This is a <b>WAMP-over-WebSocket</b> transport endpoint on your <a href="http://crossbar.io">Crossbar.io</a> server.

      Applications talk to me speaking the <b>WAMP-over-WebSocket</b> protocol.
   </p>
   {% if wsUri %}
   <p>
      Clients connect using this WebSocket URL:
   </p>
   <center>
      <pre>{{ wsUri }}</pre>
   </center>
   {% endif %}
   <p>
      If you are a human, this is likely not what you want to do;) Try <a href="/">here</a> instead.
   </p>
   <br>
   <center style="font-size: 12px;">
      Served for {{ peer }} from Crossbar.io router worker with PID {{ workerPid }}.
   </center>
   <br>

{% endblock %}
